import numpy
from _surface import SurfaceModel

from sys import stderr
from Segger import timing
from time import clock

class Segmentation ( SurfaceModel ):

    def __init__(self, name, xform = None, open = True):

        SurfaceModel.__init__(self)

        self.name = name
        self.mask = None
        self.regions = set()    # Top level regions
        self.id_to_region = {}
        self.max_region_id = 0
        self.smoothing_level = 0
        self.rcons = None       # Leaf region contacts.
        self.seg_map = None     # Map being segmented.
        self.map_level = None   # Good contouring level.
        self.ijk_to_xyz_transform = None # use map transform if map available
        self.surface_resolution = 1     # voxels

        self.adj_graph = None   # graph with regions as nodes
        self.graph_links = "uniform" # how to compute radii of links in graph
        self.regions_scale = 1.0 # for shrinking regions in the graph

        self.oneTransparentLayer = True
        from chimera import openModels as om
        if open:
            om.add([self])
            if not xform is None:
                self.openState.xform = xform

    def point_transform(self):

        if self.seg_map:
            return self.seg_map.data.ijk_to_xyz_transform
        return self.ijk_to_xyz_transform

    def voxel_volume(self):

        t = self.point_transform()
        if t is None:
            return 1

        from numpy.linalg import det
        v = det([r[:3] for r in t])
        return v

    def calculate_mask(self):

        if self.seg_map is None:
            return None

        shape = list(self.seg_map.data.size)
        shape.reverse()
        
        from numpy import zeros, uint32
        mask = zeros(shape, uint32)

        for r in self.childless_regions():
            p = r.points()
            mask[p[:,2],p[:,1],p[:,0]] = r.rid   # set mask at points

        self.mask = mask

    def region_contacts(self, task = None):

        if self.rcons is None:
            import _segment
            if timing: t0 = clock()
            rcon = _segment.region_contacts(self.mask)
            if timing: t1 = clock()
            rcons = {}
            id2r = self.id_to_region
            for i, (r1id, r2id, ncon) in enumerate(rcon):
                r1 = id2r[r1id]
                r2 = id2r[r2id]
                c = Contact(ncon)
                if not r1 in rcons:
                    rcons[r1] = {}
                if not r2 in rcons:
                    rcons[r2] = {}
                rcons[r1][r2] = rcons[r2][r1] = c
                if task and i % 1000 == 0:
                    task.updateStatus('Calculating region contacts %d of %d' %
                                      (i, len(rcon)))
            if timing: t2 = clock()
            if self.seg_map:
                import _segment
                map = self.seg_map.full_matrix()
                ci, cf = _segment.interface_values(self.mask, map)
                id2r = self.id_to_region
                for (r1id,r2id,nv), (dmax,dsum) in zip(ci,cf):
                    c = rcons[id2r[r1id]][id2r[r2id]]
                    c.maximum_density = dmax
                    c.D = dsum
                
            if timing: t3 = clock()

            self.rcons = rcons
            cc = sum([len(t) for t in rcons.values()])/2
            print 'Computed %d contacting pairs' % cc
            if timing:
                print 'Time %.2f: contact calc %.2f sec, contact objects %.2f sec, maxima %.2f sec' % (t3-t0,t1-t0,t2-t1,t3-t2)


        return self.rcons

    def contacts_changed(self):

        self.rcons = None

    def open_map(self, open = True):

        if self.seg_map or not hasattr(self, 'map_path'):
            return None
        p = self.map_path
        from os.path import isfile
        if p and isfile(p):
            from VolumeViewer import volume_list, open_volume_file
            open_paths = set([v.data.path for v in volume_list()])
            if not p in open_paths:
                if open:
                    kw = {'model_id': self.id}
                else:
                    kw = {'open_models': open}
                maps = open_volume_file(p, **kw)
                map = maps[0]
                self.seg_map = map
                if not self.map_level is None:
                    for m in maps:
                        map.set_parameters(surface_levels = [self.map_level])
                        map.show()
                return map
        return None

    def remove_all_regions(self):

        for r in self.regions:
            r.remove_surface()
        self.regions.clear()
        self.id_to_region.clear()

    def remove_region (self, region,
                       remove_children = False,
                       remove_childless_parents = True,
                       update_surfaces = True,
                       task = None):

        self.remove_regions([region], remove_children,
                            remove_childless_parents, update_surfaces, task)

    def remove_regions (self, regions,
                       remove_children = False,
                       remove_childless_parents = True,
                       update_surfaces = True,
                       task = None):

        rset = set(regions)

        if remove_children:
            for r in regions:
                rset.update(r.all_children())

        if remove_childless_parents:
            check = rset
            while check:
                childless = set()
                for r in check:
                    p = r.preg
                    if p and not p in rset:
                        if len([c for c in p.cregs if not c in rset]) == 0:
                            childless.add(p)
                rset.update(childless)
                check = childless

        parents = set()
        for i, r in enumerate(rset):

            if task and i % 100 == 0:
                task.updateStatus('remove region %d of %d' % (i, len(rset)))

            if r in self.regions:
                self.regions.remove(r)
            del self.id_to_region[r.rid]
            r.remove_surface()

            if len(r.cregs) == 0:
                p = r.points()
                if not p is None:
                    self.mask[p[:,2],p[:,1],p[:,0]] = 0   # zero mask at points
                    self.contacts_changed()

            for c in r.cregs:
                if not c in rset:
                    c.preg = None
                    self.regions.add(c)
                    if update_surfaces:
                        c.make_surface()

            p = r.preg
            if p and not p in rset:
                parents.add(p)

        for p in parents:
            p.remove_children([c for c in p.cregs if c in rset],
                              update_surfaces = update_surfaces)


    def remove_small_regions ( self, minRegionSize, task = None ) :

        rcons = self.region_contacts(task)
        regions = SmallConnectedRegions ( self.childless_regions(),
                                          rcons, minRegionSize, task )
        self.remove_regions(regions, task = task)

        print 'Removed %d regions smaller than %d voxels' % (len(regions), minRegionSize)

    def join_regions ( self, regs, max_point = None, color = None ) :

        rid = self.max_region_id + 1
        self.max_region_id += 1

        if max_point is None:
            from numpy import sum, float32, int32
            ave = sum([r.max_point for r in regs], axis=0).astype(float32) / len(regs)
            max_point = ave.astype(int32)

        nreg = Region ( self, rid, children = regs, max_point = max_point )

        if color is None:
            lr = max([(r.point_count(),r) for r in regs])[1]
            color = lr.color
        nreg.color = color

        return nreg

    def calculate_local_maxima_regions ( self, mm, thrD, task = None ) :

        self.remove_all_regions()

        self.adj_graph = None
        self.seg_map = mm

        map = mm.data.full_matrix()
        import _segment
        if timing: t0 = clock()
        if task:
            task.updateStatus('Computing watershed regions for %s' % mm.name)
        self.mask = _segment.regions_around_local_maxima(map, thrD)
        self.map_level = thrD
        if timing: t1 = clock()
        if task:
            task.updateStatus('Computing region points')
        rpoints = _segment.region_index_lists(self.mask)
        if timing: t2 = clock()
        max_points = maxima(map, rpoints, task)
        if timing: t3 = clock()
        regions = []
        n = len(rpoints)
        for r in range(n):
            if task and r % 1000 == 0:
                task.updateStatus('Created regions %d of %d' % (r, n))
            reg = Region ( self, r+1, rpoints[r], max_points[r] )
            regions.append(reg)
        
        self.regions = set(regions)
        self.id_to_region = dict([(r.rid, r) for r in regions])
        if timing: t4 = clock()

        np = sum([r.point_count() for r in regions])
        print 'Calculated %d watershed regions for %d points' % (len(regions), np)
        if timing:
            print 'Time %.2f: watershed mask %.2f, region indices %.2f, region maxima %.2f, region objects %.2f' % (t4-t0, t1-t0, t2-t1, t3-t2, t4-t3)

        return regions


    def smooth_and_group ( self, steps, sdev, min_reg = 1, task = None ) :

        if timing: t0 = clock()

        dmap = self.seg_map

        from numpy import single as floatc
        sm_mat = dmap.data.full_matrix().astype(floatc)

        rlist = []
        for iti in range ( steps ) :

            if task:
                task.updateStatus('Smoothing step %d of %d' % (iti + 1, steps))
            slev = (iti+1)*sdev

            if timing: t1 = clock()
            ijk_sdev = (slev, slev, slev)
            from VolumeFilter import gaussian
            sm = gaussian.gaussian_convolution (sm_mat, ijk_sdev, task = task)
            if timing: t2 = clock()

            rlist = self.group_by_tracking_maxima ( sm, task )
            for r in rlist:
                r.smoothing_level = slev
                
            if timing: t3 = clock()
            num_regs = len(self.regions)

            print "Smoothing width %d voxels, %d regions" % (slev, num_regs)
            if timing: 
                print 'Time %.2f: smooth %.2f, group %.2f' % (t3-t1, t2-t1, t3-t2)

            if num_regs <= min_reg : break
            
        self.smoothing_level = steps  * sdev
        return rlist


    def group_by_tracking_maxima ( self, m, task = None ) :

        rlist = list(self.regions)
        pos = numpy.array([r.max_point for r in rlist], numpy.intc)
        import _segment
        _segment.find_local_maxima(m, pos)

        rm = {}
        for r, p in zip(rlist, pos):
            p = tuple(p)
            if p in rm:
                rm[p].append(r)
            else:
                rm[p] = [r]
        groups = [(p,g) for p,g in rm.items() if len(g) >= 2]

        rlist = []
        for i, (p,g) in enumerate(groups):
            if task and i % 100 == 0:
                s = "Grouping %d of %d" % ( i, len( groups ) )
                task.updateStatus ( s )
            r = self.join_regions ( g, p )
            rlist.append(r)
        return rlist


    def group_connected ( self, regions, min_contact ) :

        # Limit connection map to specified regions.
        cons = group_contacts(self.region_contacts(), regions)

        # Dictionary mapping region to set of connected regions.
        conr = ConnectedRegions ( regions, cons, min_contact )

        # List of sets of connected regions.
        rsets = dict([(id(cr), cr) for r, cr in conr.items()]).values()

        nregs = []
        for rset in rsets :
            if len(rset) > 1:
                r = self.join_regions ( tuple(rset) )
                nregs.append(r)

        print "Created %d connected regions" % len(nregs)



    def ungroup_regions ( self, regs, task = None ):

        rlist = []
        rr = []
        for r in regs :
            if len(r.cregs) == 0 :
                rlist.append(r)
            else :
                rlist.extend(r.cregs)
                rr.append(r)
        self.remove_regions(rr, task = task)

        print 'Ungrouped %d regions into %d regions' % (len(regs), len(rlist))
        return rlist

    def display_regions(self, style = 'Voxel_Surfaces',
                        max_reg = None, task = None,
                        bForce=False):

        dmap = None
        init_mat = None
        rlist = list(self.regions)
        rlist.sort(lambda r1,r2: cmp(r2.point_count(),r1.point_count()))
        
        for i, reg in enumerate(rlist):

            if not max_reg is None and i >= max_reg :
                reg.remove_surface()
                continue

            if (not bForce) and reg.surface_piece:
                continue

            vt = None

            if style == 'Voxel_Surfaces' :
                reg.make_surface (None,None,self.regions_scale,bForce)

            elif style == 'Density_Maxima' :
                rpts = numpy.array ( [r.max_point for r in r.childless_regions()], numpy.float32 )
                import MultiScale
                vt = MultiScale.surface.surface_points ( rpts, 1.0, 0.1, .25, 5 )
                import _contour

                _contour.affine_transform_vertices ( vt[0], self.point_transform() )
                reg.make_surface ( vt[0], vt[1] )

            elif style == 'Iso_Surfaces' :

                if dmap == None:
                    dmap = self.seg_map.writable_copy ( require_copy=True )
                    init_mat = dmap.data.full_matrix().copy()

                tpoints = reg.map_points()
                vt = None
                print " - reg %d, %d voxels" % (reg.rid, len(tpoints))
                import VolumeData
                sg = VolumeData.zone_masked_grid_data( dmap.data, tpoints, dmap.data.step[0] )
                m = sg.full_matrix()

                cmatrix = dmap.data.full_matrix()
                cmatrix[:,:,:] = m[:,:,:]
                dmap.surface_levels = [ dmap.surface_levels[0] ]
                dmap.region = ( dmap.region[0], dmap.region[1], [1,1,1] )
                ro = VolumeViewer.volume.Rendering_Options()
                #ro.surface_smoothing = True
                #ro.smoothing_factor = .2
                #ro.smoothing_iterations = 5
                dmap.update_surface ( False, ro )

                surf_sp = None
                surf_sp0 = None
                for sp in dmap.surfacePieces :
                    v, t = sp.geometry
                    #print "- %d vertices, %d tris" % (len(v), len(t))
                    if len(v) == 8 and len(t) == 12 :
                        continue
                    if len(v) == 0 and len(t) == 0 :
                        surf_sp0 = sp
                    else :
                        surf_sp = sp
 
                try : vt = surf_sp.geometry
                except : vt = surf_sp0.geometry

                cmatrix[:,:,:] = init_mat[:,:,:]
                dmap.data.values_changed()
                reg.make_surface ( vt[0], vt[1] )


            if task:
                task.updateStatus('Making surface for region %d of %d'
                                  % (i, len(rlist)))


        if dmap : dmap.close()


    def color_density(self, regions = None):

        d = self.seg_map
        if d is None:
            return

        if not regions is None and len(regions) == 0:
            for p in d.surface_piece_list:
                p.vertexColors = None
            return
        
        m = self.mask
        if m is None:
            return

        ijk_to_xyz = self.point_transform()
        if ijk_to_xyz is None:
            return
        import Matrix
        xyz_to_ijk = Matrix.invert_matrix(ijk_to_xyz)

        cmap = self.region_colors(regions)
        offset = max(d.data.step)
        for p in d.surface_piece_list:
            v, t = p.geometry
            n = p.normals
            cp = v - offset*n
            import _interpolate as i
            rid, outside = i.interpolate_volume_data(cp, xyz_to_ijk, m, 'nearest')
            rid = rid.astype(numpy.uint32)    # Interpolation gives float32
            colors = cmap[rid,:]
            p.vertexColors = colors

    def region_colors(self, regions = None):

        if regions is None:
            regions = self.regions

        nid = self.max_region_id + 1
        from numpy import ones, float32
        colors = ones((nid,4), float32)
        for r in regions:
            for c in r.all_regions():
                colors[c.rid] = r.color
        return colors
            
    def all_regions ( self ):

        return self.id_to_region.values()

    def childless_regions ( self ):

        rlist = [r for r in self.id_to_region.values() if len(r.cregs) == 0]
        return rlist

    def selected_regions ( self ) :

        import Surface
        sregs = self.regions
        regions = [p.region for p in Surface.selected_surface_pieces()
                   if hasattr(p, 'region') and p.region.segmentation is self]
        return regions

    def change_surface_resolution(self, res, task = None):

        if res == self.surface_resolution:
            return
        self.surface_resolution = res
        for i, r in enumerate(self.regions):
            if task and i % 100 == 0:
                task.updateStatus('Redisplaying surface %d of %d' %
                                  (i, len(self.regions)))
            sp = r.surface_piece
            if sp:
                display = sp.display
                color = sp.color
                r.remove_surface()
                sp = r.make_surface()
                sp.display = display
                sp.color = color

    def close(self):

        from chimera import openModels as om
        om.close ( [self] )

        if self.adj_graph :
            self.adj_graph.close()

class Region:

    def __init__( self, segmentation, rid, rpoints = None,
                  max_point = None, children = None ) :

        self.segmentation = segmentation
        self.rid = rid
        segmentation.id_to_region[rid] = self
        segmentation.regions.add(self)
        segmentation.max_region_id = max(rid, segmentation.max_region_id)
        if children is None:
            children = []  # Don't use [] as default since appends change it.
        self.cregs = children   # Child regions
        self.preg = None        # Parent region
        self.smoothing_level = 0

        self.rpoints = None      # Numpy array of grid indices
        self.npoints = None
        self.max_point = None     # Position of local maximum
        self.surface_piece = None       # Displayed surface.

        self.color = random_color()          # Surface color
        self.placed = False

        if children:
            for reg in children :
                reg.preg = self
                segmentation.regions.remove(reg)
            self.max_point = max_point
        elif not rpoints is None:
            self.set_region_points(rpoints, max_point)

    def set_region_points ( self, rpoints, max_point ):

        self.rpoints = rpoints
        self.npoints = len(rpoints)

        if not max_point is None:
            max_point = tuple(max_point)
        self.max_point = max_point

    def points ( self ):

        if self.cregs:
            cpts = [r.points() for r in self.childless_regions()]
            p = numpy.concatenate(cpts) if cpts else None
        else:
            p = self.rpoints
        if p is None:
            p = numpy.empty((0,3), numpy.float32)
        return p

    def map_points ( self ):

        tpoints = numpy.array(self.points(), numpy.float32)
        tf = self.segmentation.point_transform()
        import _contour
        _contour.affine_transform_vertices ( tpoints, tf )
        return tpoints


    def center_of_points ( self, transform = True ) :

        plists = [r.points() for r in self.childless_regions()]
        s = numpy.sum([numpy.sum(plist, axis=0) for plist in plists], axis=0)
        com = s.astype(numpy.float32) / self.point_count()
        if transform:
            tf = self.segmentation.point_transform()
            import _contour
            _contour.affine_transform_vertices ( com.reshape((1,3)), tf )
            
        return com


    def region_radius ( self ) :

        from numpy import sqrt, max, sum, square
        rmax = 0
        c = self.center_of_points()
        plists = [r.map_points() for r in self.childless_regions()]
        for p in plists:
            r = sqrt ( max ( sum ( square (p - c), 1 ) ) )
            rmax = max(rmax, r)
        return rmax


    def enclosed_volume ( self ) :

        return self.point_count() * self.segmentation.voxel_volume()

    def set_color ( self, color ):

        self.color = color
        if self.surface_piece:
            self.surface_piece.color = color

    def surface ( self ) :

        return self.surface_piece

    def has_surface ( self ) :

        sp = self.surface_piece
        return sp and not sp.__destroyed__

    # Does this region or any parent have a displayed surface?
    def visible ( self ) :

        sp = self.surface_piece
        shown = sp and not sp.__destroyed__ and sp.display
        if not shown and self.preg:
            shown = self.preg.visible()
        return shown

    def make_surface ( self, vertices = None, triangles = None,
                       scale=1.0, bForce=False ):

        if (not bForce) and self.surface_piece:
            return self.surface_piece

        self.remove_surface(including_children = True)

        if vertices is None:
            seg = self.segmentation
            from MultiScale.surface import surface_points
            vertices, triangles, normals = \
                surface_points ( self.points(),
                                 resolution = seg.surface_resolution,
                                 density_threshold = 0.1, 
                                 smoothing_factor = .25,
                                 smoothing_iterations = 5 )
            import _contour
            _contour.affine_transform_vertices ( vertices, seg.point_transform() )

            import numpy
            if numpy.fabs(scale-1.0) > 0.01 :
                com = self.center_of_points ()
                t_0_com = ( (1.0,0.0,0.0,-com[0]),
                            (0.0,1.0,0.0,-com[1]),
                            (0.0,0.0,1.0,-com[2]) )
                t_to_com = ( (1.0,0.0,0.0,com[0]),
                            (0.0,1.0,0.0,com[1]),
                            (0.0,0.0,1.0,com[2]) )
                t_scale = ( (scale,0.0,0.0,0.0),
                            (0.0,scale,0.0,0.0),
                            (0.0,0.0,scale,0.0) )
                import Matrix
                tf = Matrix.multiply_matrices( t_scale, t_0_com )
                tf = Matrix.multiply_matrices( t_to_com, tf )

                _contour.affine_transform_vertices ( vertices, tf )

        nsp = self.segmentation.addPiece ( vertices, triangles, self.color )
        nsp.oslName = str(self.rid)
        nsp.region = self
        self.surface_piece = nsp
        return nsp

    def remove_surface ( self, including_children = False ):

        p = self.surface_piece
        if p:
            p.model.removePiece ( p )
        self.surface_piece = None

        if including_children:
            for c in self.cregs:
                c.remove_surface(including_children)

    # Contacting top regions.
    def contacting_regions ( self ):

        rset = set()
        rcons = self.segmentation.region_contacts()
        for r in self.childless_regions():
            rset.update(rcons[r].keys())
        cr = set()
        for r in rset:
            cr.add(r.top_parent())
        return tuple(cr)

    def parents ( self ) :

        rlist = []
        r = self
        while r.preg :
            r = r.preg
            rlist.append(r)
        return rlist

    def top_parent ( self ) :

        r = self
        while r.preg :
            r = r.preg
        return r

    def all_regions ( self ):

        rlist = [self]
        for c in self.cregs:
            rlist.extend(c.all_regions())
        return rlist

    def in_group ( self ):

        return not self.preg is None

    def has_children ( self ):

        return len(self.cregs) > 0

    def all_children ( self ):

        rlist = list(self.cregs)
        for c in self.cregs:
            rlist.extend(c.all_children())
        return rlist

    def childless_regions ( self ):

        if self.cregs:
            rlist = []
            for c in self.cregs:
                rlist.extend(c.childless_regions())
        else:
            rlist = [self]
        return rlist

    def point_count ( self ):

        np = self.npoints
        if not np is None:
            c = np
        elif self.cregs:
            self.npoints = c = sum([r.point_count() for r in self.cregs])
        else:
            p = self.points()
            self.npoints = c = 0 if p is None else len(p)
        return c

    def remove_children ( self, cregs, update_surfaces = True):

        for c in cregs:
            self.cregs.remove(c)
        self.children_changed(update_surfaces)

    def children_changed(self, update_surfaces = True):

        # Clear cached point count.
        self.npoints = None

        # Update displayed surface.
        if self.surface_piece:
            self.remove_surface()
            if update_surfaces and len(self.cregs) > 0:
                self.make_surface()

        # Update parent
        p = self.preg
        if p:
            p.children_changed(update_surfaces)

class Contact:

    def __init__ (self, ncontact):

        self.N = ncontact               # number of voxel pairs
        self.D = 0.0                    # divide by 2*N to get average
        self.maximum_density = None


def group_contacts ( rcons, limit_regions = None, task = None ) :

    if not limit_regions is None:
        limit_regions = set(limit_regions)

    # Compute top parents of leaf regions.
    top = dict([(r,r.top_parent()) for r in rcons])

    # Compute contacts of top region groups from leaf contacts.
    cons = {}
    for i, (r1, r1cons) in enumerate(rcons.items()) :
        r1_parent = top[r1]
        if limit_regions is None or r1_parent in limit_regions:
            if task and i % 1000 == 0:
                task.updateStatus('group contacts %d of %d' % (i, len(rcons)))
            r1pcons = cons.setdefault(r1_parent, {})
            for r2, o in r1cons.iteritems() :
                r2_parent = top[r2]
                if r2_parent != r1_parent:
                    if limit_regions is None or r2_parent in limit_regions:
                        if r2_parent in r1pcons:
                            no = r1pcons[r2_parent]
                            no.D += o.D
                            no.N += o.N
                            if (no.maximum_density == None or
                                no.maximum_density < o.maximum_density) :
                                no.maximum_density = o.maximum_density
                        else :
                            r1pcons[r2_parent] = Contact(o.N)

    return cons



def GroupedRegions ( regions, rset = None ):

    if rset is None:
        rset = set ( regions )
    else:
        rset.update( regions )
        
    for r in regions:
        GroupedRegions ( r.cregs, rset )

    return rset

        
def SmallRegions ( regions, minRegionSize ) :

    sreg = [r for r in regions if r.point_count() < minRegionSize ]
    return sreg


def SmallConnectedRegions ( regions, rcons, minRegionSize, task = None ) :

    conr = ConnectedRegions ( regions, rcons, task=task )
    csize = {}
    for rset in conr.values():
        s = id(rset)
        if not s in csize:
            csize[s] = sum([c.point_count() for c in rset])
    sireg = [r for r in regions if csize[id(conr[r])] < minRegionSize]
    return sireg


# Calculate dictionary mapping region to set of connected regions.
def ConnectedRegions ( regions, rcons, min_contact = None, task = None ) :

    conr = {}
    cc = 0
    for r in regions:
        if task and cc % 100 == 0:
            task.updateStatus('%d connected region sets' % cc)
        if not r in conr:
            conr[r] = s = set([r])
            bndry = [r]
            cc += 1
            while bndry :
                reg = bndry.pop()
                if reg in rcons:
                    for cr in rcons[reg]:
                        if (min_contact is None or
                            rcons[reg][cr].N >= min_contact):
                            if not cr in conr:
                                s.add(cr)
                                conr[cr] = s
                                bndry.append(cr)
    return conr


def TopParentRegions ( regions ) :

    parents = set()
    for r in regions:
        parents.add ( r.top_parent () )
    return parents

def all_regions ( regions ) :

    rlist = []
    for r in regions:
        rlist.extend(r.all_regions())
    return rlist

def maxima(map, rplist, task = None):

    mp = []
    n = len(rplist)
    for i, rpoints in enumerate(rplist):
        if task and i % 1000 == 0:
            task.updateStatus('Computing region maxima %d of %d' % (i, n))
        rvals = map[(rpoints[:,2], rpoints[:,1], rpoints[:,0])]
        mi = numpy.argmax(rvals)
        mp.append(rpoints[mi])
    return mp


def region_bounds(regions):

    rlist = []
    for r in regions:
        rlist.extend(r.childless_regions())
    ijk_min = numpy.min([r.points().min(axis=0) for r in rlist], axis=0)
    ijk_max = numpy.max([r.points().max(axis=0) for r in rlist], axis=0)
    return ijk_min, ijk_max


def random_color(avoid_rgba = None, minimum_rgba_distance = 0.2):

    from random import random as rand
    while True:
        c = ( rand()*0.7, rand()*0.7, rand()*0.7, 1.0 )
        if avoid_rgba is None:
            break
        d = sum([(x0-x1)*(x0-x1) for x0,x1 in zip(c, avoid_rgba)])
        if d > minimum_rgba_distance * minimum_rgba_distance:
            break
    return c
    

def segmentations():

    from chimera import openModels as om
    slist = om.list(modelTypes = [Segmentation])
    return slist


#
# Group first regions with the smallest drop from one region maximum to the
# contact interface maximum.
#
def group_by_contacts(smod, task = None):

    map = smod.seg_map.full_matrix()

    if task:
        task.updateStatus('Computing interface maxima')
    import _segment
    ci, cf = _segment.interface_values(smod.mask, map)

    if task:
        task.updateStatus('Computing drops')

    for r in smod.all_regions():
        r.color_level = len(smod.regions)

    # Cache top parent, optimization for high grouping depth.
    tp = {}
    for t in smod.regions:
        for c in t.childless_regions():
            tp[c] = t

    id2r = smod.id_to_region
    rc = [(id2r[id1],id2r[id2],dmax) for (id1,id2,nv),(dmax,dsum) in zip(ci,cf)]

    rmax = region_maxima(smod.regions, map)
    drc = [(min(rmax[tp[r1]],rmax[tp[r2]])-cd,r1,r2,cd) for r1,r2,cd in rc]
    drc.sort()
    drc.reverse()

    cc = 0
    rcnt = 0
    drcr = []
    dr = None
    while drc:
        d, r1, r2, cd = drc.pop()
        dl = min(rmax[tp[r1]],rmax[tp[r2]])-cd
        if dl != d:
            # Drop changed due to region merging.
            dr = dl if dr is None else min(dl, dr)
            drcr.append((dl,r1,r2,cd))
            continue
        elif not dr is None and (d > dr or len(drc) == 0):
            # Resort drops.
            drcr.append((dl,r1,r2,cd))
            drc = [(min(rmax[tp[r1]],rmax[tp[r2]])-cd,r1,r2,cd)
                   for d,r1,r2,cd in drc + drcr]
            drc.sort()
            drc.reverse()
            drcr = []
            dr = None
            rcnt += 1
            continue
        cc += 1
        if task and cc % 100 == 0:
            task.updateStatus('Contact %d of %d' % (cc, len(ci)))
        p1 = collapse_chain(tp, r1)
        p2 = collapse_chain(tp, r2)
        if p1 != p2:
            rj = smod.join_regions((p1,p2))
            rj.color_level = len(smod.regions)
            tp[p1] = tp[p2] = rj
            rmax[rj] = max((rmax[p1], rmax[p2]))
    print 'max group depth', maximum_group_depth(smod.regions)
    print 'resorted', rcnt

def connected_subsets(connections):

    subset = {}
    pairs = []
    for e1, e2 in connections:
        s1 = subset.get(e1)
        if s1 is None:
            s1 = subset[e1] = [e1]
        s2 = subset.get(e2)
        if s2 is None:
            s2 = subset[e2] = [e2]
        if not s1 is s2:
            if len(s2) > len(s1):
                t = e1,s1; e1,s1 = e2,s2; e2,s2 = t     # Swap
            s1.extend(s2)
            for e in s2:
                subset[e] = s1
            pairs.append((e1, e2))
    subsets = dict([(id(s),s) for s in subset.values()]).values()
    return subsets, pairs

def region_maxima(regions, map, rmax = None):

    if rmax is None:
        rmax = {}
    for r in regions:
        if r.cregs:
            region_maxima([c for c in r.cregs if not c in rmax], map, rmax)
            rmax[r] = max([rmax[c] for c in r.cregs])
        else:
            rmax[r] = map[r.max_point[2],r.max_point[1],r.max_point[0]]
    return rmax

def maximum_group_depth(regions):

    dmax = 0
    rd = [(r,0) for r in regions]
    while rd:
        r,d = rd.pop()
        if r.cregs:
            rd.extend([(r,d+1) for r in r.cregs])
        elif d > dmax:
            dmax = d
    return dmax

def collapse_chain(d, e):
    v = d[e]
    if v in d and v != e:
        while v in d and d[v] != v:
            v = d[v]
        while e != v:
            n = d[e]
            d[e] = v
            e = n
    return v

def regions_radius ( regions ) :

    from numpy import sqrt, max, sum, square, zeros, float32
    s = zeros((3,), float32)
    n = 0
    for reg in regions:
        for r in reg.childless_regions():
            p = r.map_points()
            s += sum(p, axis=0)
            n += len(p)
    c = s / n
    rmax = 0
    for reg in regions:
        for r in reg.childless_regions():
            p = r.map_points()
            radius = sqrt ( max ( sum ( square (p - c), axis = 1 ) ) )
            if radius > rmax:
                rmax = radius
    return rmax
